# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- Test model
# :Created:   sab 29 ott 2016 00:21:14 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

# This is adapted from
# http://docs.sqlalchemy.org/en/latest/_modules/examples/generic_associations/generic_fk.html

import sys

from sqlalchemy import create_engine, Integer, Column, String, and_, event, tuple_
from sqlalchemy.ext.declarative import as_declarative, declared_attr
from sqlalchemy.ext.hybrid import Comparator, hybrid_property
from sqlalchemy.orm import relationship, foreign, remote, backref, object_mapper


class GenericFKComparator(Comparator):
    def __eq__(self, other):
        mapper = object_mapper(other)
        pkeyv = mapper.primary_key_from_instance(other)
        return self.__clause_element__() == tuple_(other.__class__.__name__, pkeyv[0])


@as_declarative()
class Base(object):
    "Base class which provides automated table name and surrogate primary key column."

    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()
    id = Column(Integer, primary_key=True)


class Address(Base):
    """The Address class.

    This represents all address records in a single table. An instance can be attached to any
    other arbitrary entity, referencing it thru its class name and primary key.
    """

    street = Column(String)
    city = Column(String)
    zip = Column(String)

    object_kind = Column(String)
    """Refers to the type of parent."""

    object_id = Column(Integer)
    """Refers to the primary key of the parent.

    This could refer to any table.
    """

    @hybrid_property
    def related_object(self):
        "Access to the related object choosing the appropriate relationship."
        if self.object_kind:
            return getattr(self, "_%s" % self.object_kind)

    @related_object.setter
    def related_object(self, object):
        "Set the related object using the appropriate relationship."
        object_kind = object.__class__.__name__
        setattr(self, "_%s" % object_kind, object)

    @related_object.expression
    def related_object(klass):
        return tuple_(klass.object_kind, klass.object_id)

    @related_object.comparator
    def related_object(klass):
        return GenericFKComparator(tuple_(klass.object_kind, klass.object_id))


class HasAddresses(object):
    """HasAddresses mixin, creates a relationship to
    the address_association table for each parent.
    """


@event.listens_for(HasAddresses, "mapper_configured", propagate=True)
def setup_listener(mapper, class_):
    name = class_.__name__
    object_kind = name
    class_.addresses = relationship(
        Address, primaryjoin=and_(
            class_.id == foreign(remote(Address.object_id)),
            Address.object_kind == object_kind
        ),
        backref=backref("_%s" % object_kind,
                        primaryjoin=remote(class_.id) == foreign(Address.object_id)
        ))

    @event.listens_for(class_.addresses, "append")
    def append_address(target, value, initiator):
        value.object_kind = object_kind


class Customer(HasAddresses, Base):
    name = Column(String)


class Supplier(HasAddresses, Base):
    company_name = Column(String)


url = 'sqlite:////tmp/testdbloady.sqlite'
e = create_engine(url)

if len(sys.argv) == 2:
    if sys.argv[1] == 'setup':
        Base.metadata.create_all(e)
    else:
        assert e.execute("SELECT count(*)"
                         " FROM supplier s"
                         " JOIN address a"
                         "  ON a.object_id = s.id AND a.object_kind = 'Supplier'") \
                .fetchone() == (1,)

        assert e.execute("SELECT count(*)"
                         " FROM customer c"
                         " JOIN address a"
                         "  ON a.object_id = c.id AND a.object_kind = 'Customer'") \
                .fetchone() == (1,)



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- File model
# :Created:   dom 08 apr 2018 13:57:57 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Lele Gaifax
#


from sqlalchemy import Binary, Column, String, UnicodeText
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship


Base = declarative_base()

class Content(Base):
    __tablename__ = 'contents'

    code = Column(String(64), primary_key=True)
    binary = Column(Binary())
    script = Column(UnicodeText())
    content = Column(UnicodeText())



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- YAML based data loader
# :Created:   sab 02 gen 2016 16:16:52 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016, 2017 Lele Gaifax
#

from __future__ import unicode_literals

import sys
from logging import getLogger

from sqlalchemy import inspect
from sqlalchemy.orm import object_mapper, RelationshipProperty
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.exc import MultipleResultsFound

if sys.version_info.major >= 3:
    from itertools import zip_longest
else:
    from itertools import izip_longest as zip_longest

from . import File, SQL


logger = getLogger(__name__)

if sys.version_info.major >= 3:
    basestring = str


class Entity(object):
    """Model instances factory."""

    def __init__(self, model, key, fields=None, data=None, delete=False, loadonly=False):
        """Initialize a new factory.

        :type model: string
        :param model: the dotted full name of a mapped class
        :type key: either a string or a sequence of strings
        :param key: name(s) of the field(s) used to lookup existing instance
        :type fields: either None, a single string or a sequence of strings
        :param fields: if given, a list of field names
        :type data: either a single mapping or a list
        :param data: values used to initialize/update instances
        :type delete: boolean
        :param delete: if ``True``, existing instances will be deleted
        :type loadonly: boolean
        :param loadonly: if ``True``, perform only the search phase, don't update anything
        """
        self.model = model
        if isinstance(key, basestring):
            key = [key]
        self.key = key
        if isinstance(fields, basestring):
            fields = [fields]
        self.fields = fields
        if isinstance(data, dict):
            data = [data]
        self.data = data
        self.delete = delete
        self.loadonly = loadonly

    def __repr__(self):
        return "%s(model=%r, key=%r)" % (
            self.__class__.__name__,
            self.model, self.key)

    def __call__(self, session, idmap, adaptor=None):
        """Load, create or update a sequence of instances.

        :param adaptor: either None or a callable
        :rtype: an iterator over loaded/created/referenced instances
        """

        instances = self.data
        if instances is None:
            return

        for data in instances:
            instance = Instance(self, data, self.fields, idmap, adaptor)
            if self.loadonly:
                for i in instance(session, self.delete, self.loadonly):
                    yield i
            else:
                yield instance(session, self.delete, self.loadonly)


class Instance(object):
    """A single model instance."""

    def __init__(self, entity, data, fields, idmap, adaptor):
        self.entity = entity
        self.data = data
        self.fields = fields
        self.idmap = idmap
        self.adaptor = adaptor
        self.instance = None
        self.created = False

    def __call__(self, session, delete=False, loadonly=False):
        "Load an existing instance, create a new one or delete it if it exists"

        if self.instance is not None:
            return self.instance

        model = self.entity.model
        key = self.entity.key

        data = self.data
        if (self.fields is not None and isinstance(data, list)
            and len(self.fields) == len(data)):
            data = {f: v for f, v in zip(self.fields, data)}

        if self.adaptor is not None:
            data = self.adaptor(self.entity.model, self.entity.key, data)

        def getvalue(key):
            value = data.get(key, None)
            if isinstance(value, SQL):
                value = value.fetch(session)
            return self.idmap.get(id(value), value)

        filter = []
        for fname in key:
            if '->' in fname:
                attrname, _, slot = fname.partition('->')
                fvalue = getvalue(attrname)[slot]
            else:
                attrname = fname
                slot = None
                fvalue = getvalue(fname)

            if (sys.version_info.major < 3
                and isinstance(fvalue, basestring)
                and not isinstance(fvalue, unicode)):
                fvalue = fvalue.decode('utf-8')

            attr = getattr(model, attrname)
            has_custom_comparator = not isinstance(attr, InstrumentedAttribute)
            is_our_instance = isinstance(fvalue, Instance)
            if has_custom_comparator or not is_our_instance:
                if slot is not None:
                    attr = attr[slot]
                if is_our_instance:
                    filter.append(attr == fvalue(session))
                else:
                    filter.append(attr == fvalue)
            else:
                instance = fvalue(session)

                mapper = object_mapper(instance)
                pkeyf = mapper.primary_key
                pkeyv = mapper.primary_key_from_instance(instance)
                pkey = {f.name: v for f, v in zip_longest(pkeyf, pkeyv)}

                for l, r in attr.property.local_remote_pairs:
                    filter.append(getattr(model, l.name) == pkey[r.name])

        q = session.query(model)
        q = q.filter(*filter)

        if loadonly:
            return q

        try:
            obj = q.one_or_none()
        except MultipleResultsFound:
            logger.critical("Multiple instances found for %r with data %r filtered by %r",
                            model, data, ' AND '.join(str(f) for f in filter))
            raise

        if delete:
            if obj is not None:
                session.delete(obj)
            return obj

        if obj is None:
            # Create a new one
            obj = model()
            session.add(obj)
            self.created = True

        self.idmap[id(self.data)] = self
        self.instance = obj

        # Update it
        for f, v in data.items():
            try:
                attr_prop = getattr(model, f).property
            except AttributeError:
                attr_prop = None
            # If the attribute is a relationship or a traditional property (assuming its a
            # "generic foreign key") then assign the instance, otherwise its primary key
            assign_pk = not (attr_prop is None or isinstance(attr_prop, RelationshipProperty))
            setattr(obj, f, _get_value(session, self.idmap, assign_pk, f, v))

        return obj


def _get_value(session, idmap, primary_key_only, attribute, value):
    # Extract the value to be assigned: if it is a dictionary or a list, then build a copy
    # recursively taking only the primary keys of the contained instances

    value = idmap.get(id(value), value)

    if isinstance(value, Instance):
        value = value(session)
        if primary_key_only:
            if inspect(value).pending:
                session.flush()
            mapper = object_mapper(value)
            pkeyv = mapper.primary_key_from_instance(value)
            if len(pkeyv) != 1:
                raise NotImplementedError('Unable to deal with multi column PK'
                                          ' assigning %r to "%s"' % (value, attribute))
            value = pkeyv[0]
    elif isinstance(value, File):
        value = value.read()
    elif isinstance(value, SQL):
        value = value.fetch(session)
    elif isinstance(value, dict):
        value = dict((k, _get_value(session, idmap, True, attribute, value[k])) for k in value)
    elif isinstance(value, list):
        value = [_get_value(session, idmap, True, attribute, v) for v in value]

    return value



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- Test model
# :Created:   mar 08 nov 2016 09:50:08 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

import sys

from sqlalchemy import create_engine, Column, Integer, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker


Base = declarative_base()

class Product(Base):
    __tablename__ = 'products'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    attributes = Column(JSONB)


url = 'postgresql://localhost/testdbloady'
e = create_engine(url)

if len(sys.argv) == 2:
    if sys.argv[1] == 'setup':
        Base.metadata.create_all(e)
    elif sys.argv[1] == 'test':
        smaker = sessionmaker(autoflush=False, autocommit=False, bind=e)
        session = smaker()

        all_products = session.query(Product).order_by(Product.name).all()
        assert len(all_products) == 3
        p1 = all_products[0]
        assert p1.name == 'First'
        assert p1.attributes['color'] == 'Black'
        assert p1.attributes['info']['made_by'] == 'Lele'
        p2 = all_products[1]
        assert p2.name == 'Second'
        assert p2.attributes['color'] == 'White'
        assert p2.attributes['info']['made_by'] == 'Roald'
        assert p2.attributes['info']['alt_prod_id'] == p1.id
        p3 = all_products[2]
        assert p3.name == 'Third'
        assert p3.attributes['other_two'] == [p1.id, p2.id]



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- Non-relationship attributes test
# :Created:   lun 07 nov 2016 10:40:00 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

import sys

from sqlalchemy import create_engine, Column, ForeignKey, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker


Base = declarative_base()

class Person(Base):
    __tablename__ = 'persons'

    id = Column(Integer, primary_key=True)
    firstname = Column(String(64))
    lastname = Column(String(64))


class CannedFilter(Base):
    __tablename__ = 'filters'

    id = Column(Integer, primary_key=True)
    description = Column(String(64))


class Condition(Base):
    __tablename__ = 'conditions'

    id = Column(Integer, primary_key=True)
    idfilter = Column(Integer,
                      ForeignKey('filters.id', ondelete='CASCADE'),
                      nullable=False)
    fieldname = Column(String(64))
    fieldvalue = Column(String(64))

    filter = relationship('CannedFilter')


url = 'sqlite:////tmp/testdbloady.sqlite'
e = create_engine(url)

if len(sys.argv) == 2:
    if sys.argv[1] == 'setup':
        Base.metadata.create_all(e)
    elif sys.argv[1] == 'test':
        smaker = sessionmaker(autoflush=False, autocommit=False, bind=e)
        session = smaker()

        condition = session.query(Condition).one()
        assert condition.filter.description == 'Only John Doe'
        assert condition.fieldname == 'persons.id'

        person = session.query(Person).get(condition.fieldvalue)
        assert person.lastname == 'Doe'



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- Property based attributes test
# :Created:   mer 07 giu 2017 14:15:41 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017 Lele Gaifax
#

import sys

from sqlalchemy import create_engine, Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import sessionmaker


Base = declarative_base()


class PersonTP(Base):
    __tablename__ = 'persons_tp'

    id = Column(Integer, primary_key=True)
    firstname = Column(String(64))
    lastname = Column(String(64))

    @property
    def fullname(self):
        return self.firstname + ' ' + self.lastname

    @fullname.setter
    def fullname(self, value):
        self.firstname, self.lastname = value.split()


class PersonHP(Base):
    __tablename__ = 'persons_hp'

    id = Column(Integer, primary_key=True)
    firstname = Column(String(64))
    lastname = Column(String(64))

    @hybrid_property
    def fullname(self):
        return self.firstname + ' ' + self.lastname

    @fullname.setter
    def fullname(self, value):
        self.firstname, self.lastname = value.split()


url = 'sqlite:////tmp/testdbloady.sqlite'
e = create_engine(url)

if len(sys.argv) == 2:
    if sys.argv[1] == 'setup':
        Base.metadata.create_all(e)
    elif sys.argv[1] == 'test':
        smaker = sessionmaker(autoflush=False, autocommit=False, bind=e)
        session = smaker()

        person = session.query(PersonTP).one()
        assert person.fullname == 'John Doe'
        assert person.firstname == 'John'
        assert person.lastname == 'Doe'

        person = session.query(PersonHP).one()
        assert person.fullname == 'John Doe'
        assert person.firstname == 'John'
        assert person.lastname == 'Doe'



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- File test
# :Created:   dom 08 apr 2018 13:59:56 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Lele Gaifax
#

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from metapensiero.sqlalchemy.dbloady.load import load
from model import Base, Content


def main():
    e = create_engine('sqlite:///:memory:')
    Base.metadata.create_all(e)

    smaker = sessionmaker(autoflush=False, autocommit=False, bind=e)
    session = smaker()

    load('data.yaml', session)

    content = session.query(Content).get('€')

    assert content.content == 'a' * 100
    assert content.binary == b'binary.bin'
    assert ':Copyright: © 2018' in content.script


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- Sample preload script
# :Created:   gio 14 gen 2016 11:35:38 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016, 2017 Lele Gaifax
#

import datetime, re

from ruamel import yaml

def time_constructor(loader, node):
    value = loader.construct_scalar(node)
    if value.startswith('T'):
        value = value[1:]
    parts = map(int, value.split(':'))
    return datetime.time(*parts)

yaml.add_constructor('!time', time_constructor)
yaml.add_implicit_resolver('!time', re.compile(r'^T?\d{2}:\d{2}(:\d{2})?$'), ['T'])



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- File package
# :Created:   dom 08 apr 2018 13:59:31 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2018 Lele Gaifax
#

from setuptools import setup

setup(name='testdbloady',
      version='1.0',
      py_modules=['model'])



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- Test package
# :Created:   sab 09 gen 2016 10:56:57 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016, 2017 Lele Gaifax
#

from setuptools import setup

setup(name='testdbloady',
      version='1.0',
      py_modules=['model'])



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- Test model
# :Created:   gio 14 gen 2016 11:30:16 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016, 2017 Lele Gaifax
#

import sys

from sqlalchemy import create_engine, Column, Integer, String, Time
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker


Base = declarative_base()

class Event(Base):
    __tablename__ = 'events'

    id = Column(Integer, primary_key=True)
    description = Column(String(64))
    starttime = Column(Time())
    endtime = Column(Time())


url = 'sqlite:////tmp/testdbloady.sqlite'
e = create_engine(url)

if len(sys.argv) == 2:
    if sys.argv[1] == 'setup':
        Base.metadata.create_all(e)
    elif sys.argv[1] == 'test':
        smaker = sessionmaker(autoflush=False, autocommit=False, bind=e)
        session = smaker()

        event = session.query(Event).one()
        assert event.description == 'First event'
        assert str(event.starttime) == '12:23:00'
        assert str(event.endtime) == '15:34:00'



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- Test model
# :Created:   mar 15 nov 2016 14:13:56 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

import sys

from sqlalchemy import create_engine, Column, Integer
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker


Base = declarative_base()

class Number(Base):
    __tablename__ = 'numbers'

    id = Column(Integer, primary_key=True)
    absolute = Column(Integer)


if len(sys.argv) == 2:
    url = 'sqlite:////tmp/testdbloady.sqlite'
    e = create_engine(url)

    if sys.argv[1] == 'setup':
        Base.metadata.create_all(e)
    elif sys.argv[1] == 'test':
        smaker = sessionmaker(autoflush=False, autocommit=False, bind=e)
        session = smaker()

        assert session.query(Number).get(1).absolute == 1
        assert session.query(Number).get(2).absolute == 2
        assert session.query(Number).get(3).absolute == 2



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- YAML based data loader
# :Created:   ven 01 gen 2016 16:33:36 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016, 2017, 2018 Lele Gaifax
#

from __future__ import unicode_literals

from logging import getLogger
from os.path import abspath, dirname, exists, isabs
import sys

import pkg_resources

from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import object_mapper, sessionmaker

from ruamel import yaml

from . import File, resolve_class_name
from .entity import Entity


OK, LOAD_ERROR, LOAD_EXCEPTION, PRELOAD_EXCEPTION, POSTLOAD_EXCEPTION = range(5)

logger = getLogger(__name__)


def load(fname, session, dry_run=False, delete=False, save_new_instances=False,
         adaptor=None, show_progress=False):
    """Load a single YAML file.

    :param fname: the name of the YAML file to load
    :param session: the SQLAlchemy session
    :param dry_run: whether to commit data at the end
    :param delete: whether instances shall be deleted instead of updated
    :param save_new_instances: if given, the name of the YAML file where
                               information about created instances will be written
    :param adaptor: either None or a function
    :param show_progress: whether to emit some noise as the process goes on
    :rtype: dict
    :returns: A dictionary with loaded data, keyed by (model class, key): each
              value is a tuple (primarykey, datadict)

    This will open the given file (that should contain a UTF-8 encoded
    YAML structure) and will load/update the data into the database, or
    deleted from there.

    The `adaptor` function, if specified, will be called once for each "record"
    and has the opportunity of deliberately change its content::

        user_id = 999

        def adjust_user(cls, key, data):
            if key == ['username']:
                data['username'] = data['username'] + str(user_id)
                data['user_id'] = user_id
            return data

        load('testdata.yaml', session, adaptor=adjust_user)

    When `delete` is ``True``, then existing instances will be deleted
    from the database instead of created/updated.

    If `save_new_instances` is given, it's a file name that will contain a YAML
    representation of the newly created instances, suitable to be used in a
    subsequent call with `delete` set to ``True``.

    When `dry_run` is ``False`` (the default) the function performs a
    ``flush()`` on the SQLAlchemy session, but does **not** commit the
    transaction.
    """

    from io import open

    if show_progress:
        from progressbar import ProgressBar, Counter, Timer, UnknownLength, DynamicMessage

        class EntityName(DynamicMessage):
            def __call__(self, progress, data):
                val = data['dynamic_messages'][self.name]
                return ' [%s]' % (val or '...loading...')

        pbar = ProgressBar(widgets=[fname, EntityName('entity'), ': ',
                                    Counter(), ' ', Timer()],
                           max_value=UnknownLength).start()
        count = 0

    stream = open(fname, 'r', encoding='utf-8')

    # Set the base directory: file paths will be considered relative
    # to the directory containing the YAML file
    File.basedir = dirname(abspath(fname))

    idmap = {}
    loader = yaml.Loader(stream)
    while loader.check_data():
        entries = loader.get_data()
        for entry in entries:
            if show_progress:
                pbar.update(entity=entry['entity'])
            model = resolve_class_name(entry['entity'])
            if 'rows' in entry:
                data = entry['rows']
            elif 'data' in entry:
                data = entry['data']
            else:
                data = []
            entity = Entity(model, entry['key'],
                            fields=entry.get('fields', None),
                            data=data,
                            delete=delete)
            for e in entity(session, idmap, adaptor):
                if show_progress:
                    count += 1
                    pbar.update(count)

            if not dry_run:
                logger.debug("Flushing changes")
                session.flush()

    if show_progress:
        pbar.finish()
        pbar.fd.write('\n')

    if save_new_instances:
        existing_new_instances = set()
        new_new_instances = {}
        if exists(save_new_instances):
            with open(save_new_instances) as f:
                new_instances = yaml.load(f)
            for i in new_instances:
                entity = resolve_class_name(i['entity'])
                keys = i['key']
                for data in i['rows'] if 'rows' in i else i['data']:
                    key = tuple(data[key] for key in keys)
                    existing_new_instances.add((entity, key))
        else:
            new_instances = []

    result = {}
    for i in idmap.values():
        key = []
        for fname in i.entity.key:
            if '->' in fname:
                attr, _, slot = fname.partition('->')
                value = getattr(i.instance, attr)[slot]
            else:
                value = getattr(i.instance, fname)
            key.append(value)
        if len(i.entity.key) == 1:
            key = key[0]
        else:
            key = tuple(key)
        mapper = object_mapper(i.instance)
        pk = mapper.primary_key_from_instance(i.instance)

        if (save_new_instances and i.created
            and (i.entity.model, tuple(pk)) not in existing_new_instances):
            entity = i.entity.model.__module__ + '.' + i.entity.model.__name__
            pknames = tuple(str(c.key) for c in mapper.primary_key)
            data = new_new_instances.setdefault((entity, pknames), [])
            data.append(dict(zip(pknames, pk)))

        if len(pk) == 1:
            pk = pk[0]
        result[(i.entity.model, key)] = pk, i.data

    if save_new_instances and new_new_instances:
        for entity, pknames in sorted(new_new_instances):
            new_instances.append(dict(entity=entity, key=list(pknames),
                                      data=new_new_instances[(entity, pknames)]))
        with open(save_new_instances, 'w') as f:
            yaml.dump(new_instances, f, default_flow_style=False, allow_unicode=True)

    return result


def workhorse(uri, dry_run, echo, quiet,
              delete, save_new_instances, preload, postload, fnames):
    "Load one or more YAML file into the database."

    engine = create_engine(uri, echo=echo)
    salogger = getattr(engine.logger, 'logger', None)
    if salogger is not None:
        for h in salogger.handlers:
            salogger.removeHandler(h)
    smaker = sessionmaker(autoflush=False, autocommit=False, bind=engine)

    session = smaker()

    if preload is not None:
        try:
            f = open(preload)
        except IOError:
            logger.error("Could not open preload script %r!" % preload)
            return PRELOAD_EXCEPTION

        logger.debug('Executing preload script %r...', preload)
        context = dict(session=session, dry_run=dry_run, fnames=fnames)
        try:
            code = compile(f.read(), preload, 'exec')
            exec(code, context)
            fnames = context['fnames']
        except Exception:
            logger.exception("Failure executing the preload script!")
            return PRELOAD_EXCEPTION
        finally:
            f.close()

    try:
        for fname in fnames:
            load(fname, session, dry_run, delete, save_new_instances,
                 show_progress=not quiet)
    except SQLAlchemyError as e:
        # PG errors are UTF-8 encoded
        emsg = str(e)
        if sys.version_info.major < 3:
            emsg = emsg.decode('utf-8')
        logger.error("Data couldn't be loaded: %s", emsg)
        return LOAD_ERROR
    except Exception:
        logger.exception("We are in trouble, unexpected error!")
        return LOAD_EXCEPTION

    if postload is not None:
        try:
            f = open(postload)
        except IOError:
            logger.error("Could not open postload script %r!" % postload)
            return POSTLOAD_EXCEPTION

        logger.debug('Executing postload script %r...', postload)
        context = dict(session=session, dry_run=dry_run, fnames=fnames)
        try:
            code = compile(f.read(), postload, 'exec')
            exec(code, context)
        except Exception:
            logger.exception("Failure executing the postload script!")
            return POSTLOAD_EXCEPTION
        finally:
            f.close()

    if not dry_run:
        if not quiet:
            logger.info("Committing changes")
        session.commit()

    return OK


def path_spec(ps):
    if isabs(ps) or not ':' in ps:
        return ps
    pkgname, subpath = ps.split(':', 1)
    return pkg_resources.resource_filename(pkgname, subpath)


def main():
    import locale, logging
    from argparse import ArgumentParser, RawDescriptionHelpFormatter

    locale.setlocale(locale.LC_ALL, '')

    parser = ArgumentParser(
        description="Load and/or update DB model instances.",
        epilog=__doc__, formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument("datafile", nargs="+", type=path_spec,
                        help="The YAML data file to load. It may be either a plain"
                        " file name, or a package relative path like"
                        " “package.name:some/file”.")
    parser.add_argument("-u", "--sqlalchemy-uri", type=str, metavar="URI",
                        help="Specify the SQLAlchemy URI.", default=None)
    parser.add_argument("-D", "--delete", default=False, action="store_true",
                        help="Delete existing instances instead of creating/"
                        "updating them. You better know what you are doing!")
    parser.add_argument("-s", "--save-new-instances", type=str, metavar='FILE',
                        help="Save new instances information into given YAML file,"
                        " preserving it's previous content.")
    parser.add_argument("-p", "--preload", type=path_spec, metavar='SCRIPT',
                        help="Execute the given Python script before loading the"
                        " data files. It may be either a plain file name or a package"
                        " relative path like “package.name:some/file”.")
    parser.add_argument("-P", "--postload", type=path_spec, metavar='SCRIPT',
                        help="Execute the given Python script after load but before"
                        " committing changes. It may be either a plain file name or a"
                        " package relative path like"
                        " “package.name:some/file”.")
    parser.add_argument("-n", "--dry-run", default=False, action="store_true",
                        help="Don't commit the changes to the database.")
    parser.add_argument("-e", "--echo", default=False, action="store_true",
                        help="Activate SA engine echo")
    parser.add_argument("-q", "--quiet", default=False, action="store_true",
                        help="Be quiet, emit only error messages")
    parser.add_argument("-d", "--debug", default=False, action="store_true",
                        help="Activate debug logging")
    if sys.version_info.major < 3:
        parser.add_argument("-w", "--unicode-warnings", default=False,
                            action="store_true",
                            help="Activate SA unicode warnings")

    args = parser.parse_args()

    logging.basicConfig(format='%(message)s',
                        level=logging.DEBUG if args.debug else logging.INFO)

    if args.sqlalchemy_uri is None:
        print("You must specify the SQLAlchemy URI, example:")
        print("  python %s -u postgresql://localhost/dbname data.yaml"
              % sys.argv[0])

        return 128

    if sys.version_info.major < 3 and args.unicode_warnings:
        import warnings
        from sqlalchemy.exc import SAWarning

        warnings.filterwarnings(
            'ignore', category=SAWarning,
            message="Unicode type received non-unicode bind param value")

    return workhorse(args.sqlalchemy_uri, args.dry_run, args.echo, args.quiet,
                     args.delete, args.save_new_instances,
                     args.preload, args.postload, args.datafile)


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- Test model
# :Created:   gio 22 ott 2015 18:07:50 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2015, 2016, 2017 Lele Gaifax
#

import sys

from sqlalchemy import create_engine, Column, ForeignKey, Integer, String
from sqlalchemy.dialects.postgresql import HSTORE
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker


Base = declarative_base()

class Product(Base):
    __tablename__ = 'products'

    id = Column(Integer, primary_key=True)
    description = Column(HSTORE)


class Attribute(Base):
    __tablename__ = 'attributes'

    id = Column(Integer, primary_key=True)
    name = Column(String(64))
    description = Column(HSTORE)


class Values(Base):
    __tablename__ = 'product_attributes'

    id = Column(Integer, primary_key=True)
    idproduct = Column(Integer,
                       ForeignKey('products.id', ondelete='CASCADE'),
                       nullable=False)
    idattribute = Column(Integer, ForeignKey('attributes.id', ondelete='CASCADE'),
                         nullable=False)
    value = Column(String(64))

    product = relationship('Product')
    attribute = relationship('Attribute')


url = 'postgresql://localhost/testdbloady'
e = create_engine(url)

if len(sys.argv) == 2:
    if sys.argv[1] == 'setup':
        Base.metadata.create_all(e)
    else:
        smaker = sessionmaker(autoflush=False, autocommit=False, bind=e)
        session = smaker()

        if sys.argv[1] == 'test_1':
            all_values = session.query(Values).order_by(Values.value).all()
            assert len(all_values) == 3
            v = all_values[0]
            assert v.product.description['en'] == 'Roadrunner cage'
            assert v.product.description['it'] == 'Gabbia per struzzi'
            assert v.attribute.description['en'] == 'Size'
            assert v.attribute.description['it'] == 'Dimensione'
            v = all_values[1]
            assert v.value == '230kg'
            assert v.attribute.description['en'] == 'Weight'
            assert v.attribute.description['it'] == 'Peso'
            v = all_values[2]
            assert v.value == 'Box'
            assert v.attribute.description['en'] == 'Aspect'
            assert v.attribute.description['it'] == 'Aspetto'
        elif sys.argv[1] == 'test_2':
            assert session.query(Product).all() == []
            assert session.query(Attribute).all() == []
            assert session.query(Values).all() == []
